/*
 * MSP430 LCR meter
 * Free for non-commercial use.
 * 
 * Contact: xyphro@gmail.com
 */
 
#include "io430.h"
#include "dac.h"
#include <intrinsics.h>                                                         /* needed for __delay_cycles macro */
#include "pga.h"
#include "sinetabs.h"

/*
  signal generation works with DAC, Timer & DAC1:
  => DAC1 is updated by Timer OUT signal
  => then DAC1IFG toggles, which triggers the DMA

  sample and hold time (RS=100 Ohm assumed, although it is close to 0):
  tsample > (100+1800 Ohm)*ln(2^13)*25pF + 800ns 
  => tsample > 1228 ns => (@4 MHz) tsample >= 4.9 Cycles, next best fit is 16 Cycles!


 observations:
 - I should avoid using the internal ADC/DAC reference as it has the tendency to 
   introduce much more noice as when using the AVCC as reference 
 */

/* count of samples per measurement-burst, which are invalid and skipped*/
#define PRECEEDING_SAMPLES 2


/* dac sample rate = 1 MHz (Period = 1us) */
int16_t           sample_sine[1200+PRECEEDING_SAMPLES];                         /* raw adc12 sample data values will be written here */
int16_t           sample_pos;                                                   /* pointer to sample_sine buffer - sampling starts at zero and stops at 1199*/
volatile bool     sample_finished;                                              /* set to true when sampling has finished */

uint8_t           current_freq = 255;


/* optimize this interrupt function! */
/* ADC interrupt... */
#pragma vector = TIMER0_A1_VECTOR
__interrupt void TIMER0_CCTL1_ISR(void)
{
  uint16_t tmp, samplevalue;
  //P4OUT |= (1 << 0);
  tmp = TA0IV;
  
  samplevalue = ADC12MEM0;
  
  TA0CCTL1 = OUTMOD_0 | CCIE; 
  TA0CCTL1 = OUTMOD_3 | CCIE;                                                   /* tweak to reset OUT bit manually, which is not possible in OUTMOD_3 :-) */ 
  
  tmp = TA0CCR1;
  tmp += 10;
  if (tmp >= (SAMPLE_COUNT*10)-1)
  {
    tmp = 1; /* wrap over to beginning... */
  }
  TA0CCR1 = tmp;

  /* write ADC value to samplebuffer */  
  sample_sine[sample_pos] = samplevalue;
  
  sample_pos++;    
  if (sample_pos >= SAMPLE_COUNT+PRECEEDING_SAMPLES)
  {
    TA0CCTL1 &= ~CCIE;                                                        /* disable CCTRL1 interrupt to avoid overwriting of last sample */
    sample_finished = true; 
  }
  //P4OUT &= ~(1 << 0);
}

void dac_switch_wave(uint8_t freq)
{
  if (current_freq != freq)
  {
    current_freq = freq;
    dac_stop_wave();
    DAC12_1CTL0 &= ~DAC12IFG; /* clear DAC12 Interrupt flag to be able to restart DMA transfer */
    dac_init();
    dac_start_wave(freq);
  }
}

/* start waveform generation */
void dac_start_wave(uint8_t freq)
{
  int16_t const *sin_tab;
  uint16_t sin_tab_length;
  
  current_freq = freq;
  
  switch (freq)
  {
  case FREQUENCY_100HZ:
    sin_tab = Sin_tab_100Hz;
    sin_tab_length = sizeof(Sin_tab_100Hz) >> 1;
    break;
  case FREQUENCY_1000HZ:
    sin_tab = Sin_tab_1kHz;
    sin_tab_length = sizeof(Sin_tab_1kHz) >> 1;
    break;
  case FREQUENCY_1000HZ_CAL:
    sin_tab = Sin_tab_1000Hz_calib;
    sin_tab_length = sizeof(Sin_tab_1000Hz_calib) >> 1;
    break;    
  case FREQUENCY_10000HZ:
  default:
    sin_tab = Sin_tab_10kHz;
    sin_tab_length = sizeof(Sin_tab_10kHz) >> 1;
    break;
  }
    
  /* program dma transfer */
  DMA0SA = (unsigned short)sin_tab;         // Source block address
  DMA0DA = (unsigned short)&DAC12_1DAT;     // Destination single address
  DMA0SZ = sin_tab_length;                  // Block size
  DMACTL0 = 26;                             // DAC12IFG trigger (= DAC12_1IFG)
  DMA0CTL = DMADT_4 + DMASRCINCR_3 + DMAEN;
                                            // Repeated single transfer,
                                            // DMA source address increment,
                                            // since DMASRCBYTE = 0, the source address
                                            // increments by two (word-word)
  
  /* setup DA converter*/
  DAC12_0DAT = 0x0000; // DAC_0 output 0V
  DAC12_0CTL0 = DAC12SREF_1 | DAC12IR | DAC12AMP_5 | DAC12ENC | DAC12LSEL_0 | DAC12DF; // 2's complement
  DAC12_1DAT = 0x0000; // DAC_0 output 0V
  DAC12_1CTL0 = DAC12SREF_1 | DAC12IFG | DAC12IR | DAC12AMP_7 | DAC12ENC | DAC12LSEL_2 | DAC12DF; // 2's complement, update on rising edge of Timer0_A5 CCR1 output
  
  /* error in MSP430F5338 datasheet chapter 30.2.4: It says TimerA0 CCR1, but TimerA1 CCR1 triggers DAC latch! */
  TA1CCTL1 = OUTMOD_3;                      // TACCR1 set/reset
  TA1CCR1 = 6;                              // TACCR1 PWM Duty Cycle
  TA1CCR0 = 10-1;                           // Clock period of TACCR0 
  
  TA0CCTL1 = OUTMOD_3 | CCIE;                // TACCR1 set/reset
  TA0CCR1 = 1+0;                             // TACCR1 PWM Duty Cycle
  TA0CCR0 = SAMPLE_COUNT*10-1;               // Clock period of TACCR0
  TA0EX0  = TAIDEX_4;                        // Divide 12 MHz timer input clock by 5 (gives a total scale of 5 with ID-divider factor of 2)
  
  ADC12CTL0 = ADC12SHT0_1 | ADC12ON;         // Sampling time, ADC12 on
  ADC12CTL1 = ADC12DIV_2 | ADC12SSEL_3 | ADC12SHP | ADC12SHS_1 | ADC12CONSEQ_2;         // Use sampling timer, 12 MHz/3 = 4 MHz ADC clock
  ADC12IE = 0x00;                           // Disable interrupt
// was SREF_1!  
  ADC12MCTL0 = ADC12SREF_0; /* select VREF+ / AVSS as reference */
  ADC12CTL0 |= ADC12ENC;
  ADC12CTL2 |= ADC12SR | ADC12REFBURST; /* reference buffer setup for 200 ksps*/
  P6SEL |= 0x01;                            // P6.0 ADC option select
  
  TA1CTL = TACLR | TASSEL_2 | MC_1;                                             
  TA0CTL = TACLR | TASSEL_2 | MC_1 | ID_1;    // SMCLK, contmode, additional divider of 2
}

/* make a specified count of measurements */
int16_t *dac_measure(void)
{
  sample_finished = false;
  
  //for (int i=0; i<(1200+PRECEEDING_SAMPLES); i++)
  //  sample_sine[i] = 0;
  
  
  //TA0CCR1 = 1+0;
  //TA0CCTL1 = OUTMOD_0 | (1 << 2);                                               /* set OUT bit to avoid false ADC triggering */
  //TA0CCTL1 = OUTMOD_3;    
  //ADC12IFG |= 0xffff;                                                           /* disable any interrupt flags */
  
  TA0CCR1 = 1+10*600;                             // TACCR1 PWM Duty Cycle
  TA0CCTL1 &= ~(CCIFG); /* reset interrupt flag */
  TA0CCTL1 |= CCIE;                /* enable capture compare interrupt*/
  
  //sample_first = 1;
  sample_pos = 0;
  __bis_SR_register(GIE);                                                       /* enable interrupt */
  while (!sample_finished)
    ;                                                                           /* wait until conversion is done*/
  __bic_SR_register(GIE);                                                       /* disable interrupt -> already done in interrupt, but just to be sure... */  
  
  return &sample_sine[PRECEEDING_SAMPLES];
}



/* stop waveform generation. Post state: Both lines = common mode voltage */
void dac_stop_wave(void)
{
  __bic_SR_register(GIE);                                                       /* disable interrupt */  
  DAC12_0CTL0 = DAC12IR | DAC12AMP_5 | DAC12ENC | DAC12LSEL_0 | DAC12DF;        /* setup DAC_0 */
  DAC12_1CTL0 = DAC12IR | DAC12AMP_5 | DAC12ENC | DAC12LSEL_0 | DAC12DF;        /* setup DAC_1 */
  DAC12_0DAT = 0x0000;                                                          /* DAC_0 => output 0V */
  DAC12_1DAT = 0x0000;                                                          /* DAC_1 => output 0V */
  DMA0CTL = 0;                                                                  /* turn off dma */
  TA0CTL = 0;                                                                   /* turn off adc trigger timer */
  TA1CTL = 0;                                                                   /* turn off dac sine generation timer */
  ADC12CTL0 = 0;                                                                /* turn off ADC*/
}



void dac_init(void)
{  
  /* turn internal 2.5V reference on */
  //ADC12CTL0 = ADC12REF2_5V | ADC12REFON;
  REFCTL0 = 0xbb ;
  
  /* setup DA converter*/
  //DAC12_0CTL0 = DAC12IR | DAC12AMP_5 | DAC12ENC | DAC12LSEL_0 | DAC12DF; // 2's complement
  //DAC12_0DAT = 0x0000; // DAC_0 output 0V
  //DAC12_1CTL0 = DAC12IR | DAC12AMP_5 | DAC12ENC | DAC12LSEL_0 | DAC12DF; // 2's complement, update on rising edge of Timer0_A5 CCR1 output
  //DAC12_1DAT = 0x0000; // DAC_0 output 0V
  

  P1SEL |= (1 << 1) | (1 << 2);
  P1DIR |= (1 << 1) | (1 << 2);
  
  P4DIR |= (1 << 0);
}

