/*
 * MSP430 LCR meter
 * Free for non-commercial use.
 * 
 * Contact: xyphro@gmail.com
 */
 
#include "io430.h"
#include "lcd.h"
#include <intrinsics.h>                                                         /* needed for __delay_cycles macro */
#include "font.h"

/*
 tiny library for nokia 3x10 display
 datasheet of display controller: http://www.myplace.nu/mp3/files/pcd8544.pdf

 reset is low active
 D/C => low => command, high = data, polled only at last rising edge
 SCE => low active
 Data is sampled on rising edge of SCLK, 8 Bit wide data, MSB first
 4 MHz max. SCLK freq.
 min. /RES width 100ns
   
 resolution: 84 x 48 pixel (14 * 8 characters)
 */

static void s_lcd_set_reset(bool high)
{
  if (high)
    P8OUT |=  (1 << 6);
  else
    P8OUT &= ~(1 << 6);
}

static inline void s_lcd_send(bool isCommand, uint8_t data, bool keep_cs_low)
{
  P8OUT &= ~(1 << 1);                                                           /* CS = low */
  if (isCommand)
    P8OUT &= ~(1 << 7);
  else
    P8OUT |=  (1 << 7);                                                         /* set data/command selector output */
  
  while (!(UCB1IFG&UCTXIFG));               // USCI_A0 TX buffer ready?
  UCB1TXBUF = data;                     // Transmit first character
  
  if (!keep_cs_low)
  {
    while ((UCB1STAT&UCBUSY));               // transmitter still busy?
    P8OUT |= (1 << 1);                                                          /* CS = high */
  }
}

static inline void s_lcd_cmd(uint8_t cmd)
{
  s_lcd_send(true, cmd, false);
}



void lcd_init(void)
{
  uint8_t tmp=0;
  
  P8SEL |= (1 << 4) | (1 << 5);                                                 /* initialize IOs */
  P8DIR |= (1 << 7) | (1 << 6) | (1 << 1);                                      /* set non-spi IOs */
  P8OUT |= (1 << 1);                                                            /* CS = high */
    
  s_lcd_set_reset(false);
  __delay_cycles(12);
  s_lcd_set_reset(true);                                                          /* generate reset pulse (spec.: 100ns min width) */
  
  UCB1CTL1 |= UCSWRST;                      // **Put state machine in reset**
  UCB1CTL0 |= UCMST+UCSYNC+UCCKPL+UCMSB;    // 3-pin, 8-bit SPI master
                                            // Clock polarity high, MSB
  UCB1CTL1 |= UCSSEL_2;                     // SMCLK
  UCB1BR0 = 3;                              // /3 ( => 4 MHz)
  UCB1BR1 = 0;                              //
  UCB1CTL1 &= ~UCSWRST;                     // **Initialize USCI state machine**

  s_lcd_cmd(0x21);
  s_lcd_cmd(0xc8);
  s_lcd_cmd(0x13);
  s_lcd_cmd(0x20);
  lcd_clear();
  s_lcd_cmd(0x09);
  s_lcd_cmd(0x08);
  __delay_cycles(12000ul*10);
  s_lcd_cmd(0x0c);
  
  lcd_gotoxy(0, 0);
}

/* y in uints of 8 pixels! */
void lcd_gotoxy(uint8_t x, uint8_t y)
{
  s_lcd_cmd(0x40 | (y & 0x07));
  s_lcd_cmd(0x80 | (x & 0x7f));
}

void lcd_clear(void)
{
  uint16_t  cnt;  
  
  lcd_gotoxy(0, 0);
  for (cnt=84*8; cnt > 0; cnt--)
  {
    s_lcd_send(false, 0x00, (cnt != 0));  
  }
}


uint16_t lcd_textwidth(const uint8_t *font, const char *ch)
{
  uint16_t width = 0;
  
  while (*ch) /* for each character in string */
  {           /* draw each character */
    width += font[(font[1 + *ch*2 + 1] << 8) | (font[1 + *ch*2])] + 1;
    ch++;
  }        
  return width;
}

void lcd_printstring(uint8_t x, uint8_t y, const uint8_t *font, const char *ch)
{
  uint8_t  font_line_height = (font[0]+7) >> 3; // font height, unit: bytes
  const uint8_t *ptr_char;
  const char     *cur_ch;
  uint8_t  line, char_width;
  
  for (line=0; line < font_line_height; line++) /* for each line-byte of string */
  {
    cur_ch = ch;
    
    lcd_gotoxy(x, y + line);
    while (*cur_ch) /* for each character in string */
    {           /* draw each character */
      ptr_char   = &font[(font[1 + *cur_ch*2 + 1] << 8) | (font[1 + *cur_ch*2])];
      char_width = *ptr_char++;
      ptr_char += line;
      
      while (char_width--)
      { /* draw each vertical line segment*/
        s_lcd_send(false, *ptr_char, false);
        ptr_char += font_line_height;
      }
      s_lcd_send(false, 0x00, false);  /* draw 1 character blank between each character */
      
      cur_ch++;
    }        
  }
}


void nokia_contrast (char contrast) 
{
  s_lcd_cmd(0x21);                              // LCD Extended Commands.
  s_lcd_cmd(0x80 | (contrast & 0x7f));          // Set LCD Vop (Contrast).
  s_lcd_cmd(0x20);                              // LCD Standard Commands, horizontal addressing mode.
}
