/*
 * MSP430 LCR meter
 * Free for non-commercial use.
 * 
 * Contact: xyphro@gmail.com
 */
 
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>

#include "io430.h"
#include "sys.h"
#include "pga.h"
#include "lcd.h"
#include "uart.h"
#include "dac.h"
#include "measure.h"
#include "key.h"

/* calibration values */
double current_sensor_cmrr = 0.000598270446;       /* linear, not dB! */
double voltage_to_current_gain = 1.00285279f;      /* V_Gain / I_GAIN */
double gain_values[8] = {   1.0f,
                            2.01923871f,
                            5.08245182f,
                           10.1771679f,
                           20.3397102f,
                           50.846054f,
                          101.686599f,
                          202.865142f    };
double calib_Rsense = 1000.0f; /* resistance of current sense resistor */

static __no_init int16_t samples_U[SAMPLE_COUNT];
static __no_init int16_t samples_I[SAMPLE_COUNT];

char str[64];
char str2[64];

bool breakoperation = false;



/*
 calibrate PGA gains. The tweezer contacts have to be open
*/
void calibrate_gain(void)
{
  uint8_t gainfac_v;
  int16_t *psamples;
  
  dac_switch_wave(FREQUENCY_1000HZ_CAL);
    
  for(gainfac_v=0; gainfac_v<8; gainfac_v++)
  {
    double amplitude_avg = 0;
    double amplitude;
    
    lcd_clear();
    lcd_printstring(0, 0, Arial_7, "Calibrating Gain" );
    sprintf(str, "%d/8", gainfac_v+1);
    lcd_printstring(0, 1, Impact_12, str);
      
    pga11x_set(gainfac_v, 0); /* 0 = voltage, 1 = current*/
    for (int i=0; i < 64; i++)
    {
      psamples = dac_measure();
      {        
        measure(psamples, FREQUENCY_1000HZ_CAL, 0, 0, 0, &amplitude, 0, 0, 0);
        amplitude_avg += amplitude;
      }
    }
    amplitude_avg /= 64;
    gain_values[gainfac_v] = amplitude_avg;
    //printf("G: %d %1.6f\n", gainfac_v, amplitude_avg);
  }
  for(int i=1; i<8; i++)
  {
    gain_values[i] /= gain_values[0];    
    //printf("%3.4f\t", gain_values[i]);
  }
  //printf("\n");

  lcd_clear();

}

/*
  calibrate relative current signal to voltage signal

  \TODO: Consider opamp output impedance!!!!!!!!!!!!
 */
void calibrate_relative_gain(void)
{
  int16_t *psamples;
  double angle_u, amplitude_u;
  double angle_i, amplitude_i;
  uint8_t over_underflow;
  
  dac_switch_wave(FREQUENCY_100HZ);
  
  /* step 1: Measure voltage signal amplitude and phase with GAIN=1 */
  lcd_clear();
  lcd_printstring(0, 0, Arial_7, "Open tweezer");
  lcd_printstring(0, 2, Arial_7, "press BTN");
  wait_key();
  lcd_clear();
  pga11x_set(0, 0); /* 0 = voltage, 1 = current*/
  psamples = dac_measure();
  measure(psamples, FREQUENCY_100HZ, 0, 0, &angle_u, &amplitude_u, &over_underflow, 0, 0);
  memcpy(samples_U, psamples, SAMPLE_COUNT*2);

  /* step 2: Measure current signal amplitude and phase with GAIN=1 */  
  lcd_printstring(0, 0, Arial_7, "Close tweezer");
  lcd_printstring(0, 2, Arial_7, "press BTN");
  wait_key();
  lcd_clear();
  pga11x_set(0, 1); /* 0 = voltage, 1 = current*/
  psamples = dac_measure();  
  measure(psamples, FREQUENCY_100HZ, 0, 0, &angle_i, &amplitude_i, &over_underflow, 0, 0);
  
  voltage_to_current_gain = amplitude_u / amplitude_i;
  
  sprintf(str, "%1.5f ", voltage_to_current_gain);
  lcd_printstring(0, 0, Arial_7, "relative gain:");
  lcd_printstring(0, 1, Arial_7, str);
  lcd_printstring(0, 2, Arial_7, "press BTN");
  
  wait_key();
  lcd_clear();
}

/*
  calibrate imbalance of subtractor. 
  This allows to improve the CMRR digitally...
  The tweezer contacts have to be => open
*/
void calibrate_cmrr(void)
{
  int i, j;
  int16_t *psamples;
  int16_t mean;
  double angle_u, amplitude_u;
  double angle_i, amplitude_i;
  double cmrr;
  uint8_t over_underflow;
  
  dac_switch_wave(FREQUENCY_100HZ);
  
  
  lcd_clear();
  lcd_printstring(0, 0, Arial_7, "Open tweezer");
  lcd_printstring(0, 2, Arial_7, "press BTN");
  wait_key();
  
  lcd_clear();
  
  /* step 1: Measure voltage signal amplitude and phase with GAIN=1 */
  pga11x_set(0, 0); /* 0 = voltage, 1 = current*/
  psamples = dac_measure();
  measure(psamples, FREQUENCY_100HZ, 0, 0, &angle_u, &amplitude_u, &over_underflow, 0, 0);
  memcpy(samples_U, psamples, SAMPLE_COUNT*2);

  /* step 2: Measure current signal amplitude and phase with GAIN=200 */  
  pga11x_set(7, 1); /* 0 = voltage, 1 = current*/
  
  memset(samples_I, 0, SAMPLE_COUNT*2);

  for (i=0; i<8; i++)
  {
    psamples = dac_measure();  
    for(j=0; j<SAMPLE_COUNT; j++)
    {
      samples_I[j] += *psamples++;
    }
  }
  for(j=0; j<SAMPLE_COUNT; j++)
  {
    samples_I[j] = (samples_I[j]+4) >> 3;
  }  
  
  measure(samples_I, FREQUENCY_100HZ, &mean, 0, &angle_i, &amplitude_i, &over_underflow, 0, 0);
  
  
  
  amplitude_i = amplitude_i / pga_gainvalues[7];

  //printf("amplitude_i: %f\n", amplitude_i);
  
  cmrr = amplitude_i / amplitude_u;  
  if (fabs(angle_u - angle_i) > 90)
  { /* signals are out of phase => negative gain */
    cmrr = -cmrr;
  }
  
  current_sensor_cmrr = cmrr;
  
  lcd_printstring(0, 0, Arial_7, "CMRR before:");
  sprintf(str, "%3.1f dB", 20*log10(cmrr));
  lcd_printstring(0, 1, Impact_12, str);
  
  
  //printf("CMRR: %f\n", cmrr);
  //printf("CMRR (log): %f\n", 20*log10(cmrr));
  
  /* check if correction factor is good :-) */
  
  // 1.) amplitude
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    samples_I[i] = samples_I[i] - (int16_t)((samples_U[i]-mean)*cmrr*pga_gainvalues[7]/pga_gainvalues[0]);
  }
  measure(samples_I, FREQUENCY_100HZ, 0, 0, &angle_i, &amplitude_i, &over_underflow, 0, 0);
  amplitude_i = amplitude_i / pga_gainvalues[7];
  cmrr = amplitude_i / amplitude_u;  

  
  lcd_printstring(0, 3, Arial_7, "CMRR after:");
  sprintf(str, "%3.1f dB", 20*log10(cmrr));
  lcd_printstring(0, 4, Impact_12, str);
  
  wait_key();
  lcd_clear();
}


#define ADC_SAFETYMARGIN (4096/100)

/* return next gain setting */
int check_pgascale(int16_t *insamples, uint8_t oldgain)
{
  int i;
  int16_t value, min, max;
  int32_t mean, amplitude;
  uint8_t prevgain, nextgain;
  int32_t nextamplitude;
  uint8_t newgain;
   
  newgain = oldgain;
  
  min = 4096;
  max = 0;
  mean = 0;
  for (i=SAMPLE_COUNT; i>0; i--)
  {
    value = *insamples++;
    if (value < min)
      min = value;
    if (value > max)
      max = value;    
    mean += value;
  }
  mean = mean / SAMPLE_COUNT;
  
  /* check amplitude distance to ADC over/underrange => carefull approach as mean is not always value 2048!*/
  
  /* measure_amplitude */
  min -= mean;
  max -= mean;
  amplitude = -min;
  if (max > amplitude)
    amplitude = max;
  
  /* make prediction of amplitudes with smaller or larger gain setting */
  prevgain  = oldgain;
  if (oldgain > 0)
    prevgain = oldgain-1;
  nextgain  = oldgain;
  if (oldgain < PGA_GAINCOUNT-1)
    nextgain = oldgain+1;  
  //prevamplitude = (amplitude*pga_gainvalues[prevgain]) / pga_gainvalues[oldgain];
  nextamplitude = (amplitude*pga_gainvalues[nextgain]) / pga_gainvalues[oldgain];
  
  if (nextamplitude < (2048-ADC_SAFETYMARGIN))
    newgain = nextgain;
  else if (amplitude > (2048-ADC_SAFETYMARGIN))/* current amplitude is too big */
    newgain = prevgain;
  
  return newgain;  
}


void measure_once(uint8_t freq, double *pRes, double *pAngle, double *pVp, double *pVq, double *pIp, double *pIq)
{
  int             i;
  static uint8_t  s_gainfac_v = 0;
  static uint8_t  s_gainfac_i = 0;
  uint8_t         gainfac_old;
  int16_t        *psamples;
  int16_t         mean_u;
  
  double          angle_U, angle_I, amplitude_U, amplitude_I;
  
  dac_switch_wave(freq);
  
  /* measure voltage signal */
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    samples_U[i] = 0;
  }
  do
  {
    breakoperation |= (get_key() != 0); if (breakoperation) return;
    gainfac_old = s_gainfac_v;
    pga11x_set(s_gainfac_v, 0); /* 0 = voltage, 1 = current*/
    psamples = dac_measure();
    s_gainfac_v = check_pgascale(psamples, s_gainfac_v);
    //printf("Gainfac U: %d\n", s_gainfac_v);
  } while (s_gainfac_v != gainfac_old);

  for(int j=0; j<8; j++)
  {
    breakoperation |= (get_key() != 0); if (breakoperation) return;
    if (j > 0)
    {
      psamples = dac_measure();
    }
    for (i=0; i < SAMPLE_COUNT; i++)
    {
      samples_U[i] += psamples[i];
    }
  }
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    samples_U[i] >>= 3;
  }

  measure(samples_U, freq, &mean_u, 0, &angle_U, &amplitude_U, 0, pVp, pVq);

  //memcpy(samples_U, psamples, SAMPLE_COUNT*2); /* store samples for improving CMRR of current measurement */
  
  
  /* measure current signal */
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    samples_I[i] = 0;
  }
  do
  {
    breakoperation |= (get_key() != 0); if (breakoperation) return;
    gainfac_old = s_gainfac_i;
    pga11x_set(s_gainfac_i, 1); /* 0 = voltage, 1 = current*/
    psamples = dac_measure();    
    s_gainfac_i = check_pgascale(psamples, s_gainfac_i);
    //printf("Gainfac I: %d\n", s_gainfac_i);
  } while (s_gainfac_i != gainfac_old);
  for(int j=0; j<8; j++)
  {
    breakoperation |= (get_key() != 0); if (breakoperation) return;
    if (j > 0)
    {
      psamples = dac_measure();
    }
    for (i=0; i < SAMPLE_COUNT; i++)
    {
      samples_I[i] += psamples[i];
    }
  }
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    samples_I[i] >>= 3;
  }
  psamples = samples_I;
     
  /* apply CMRR calibration */
  for (i=0; i < SAMPLE_COUNT; i++)
  {
    psamples[i] = psamples[i] - (int16_t)((samples_U[i]-mean_u)*current_sensor_cmrr*pga_gainvalues[s_gainfac_i]/pga_gainvalues[s_gainfac_v]);
  }  
  measure(psamples, freq, 0, 0, &angle_I, &amplitude_I, 0, pIp, pIq);

  if (pVp)  *pVp /= gain_values[s_gainfac_v];
  if (pVq)  *pVq /= gain_values[s_gainfac_v];
  if (pIp)  *pIp /= gain_values[s_gainfac_i] * calib_Rsense; /* negation due to phaseshift in hardware circuit! */
  if (pIq)  *pIq /= gain_values[s_gainfac_i] * calib_Rsense;
  
  amplitude_U /= gain_values[s_gainfac_v];
  amplitude_I /= gain_values[s_gainfac_i];
  
  //amplitude_I *= voltage_to_current_gain;
  
  if (pAngle)
  {
    *pAngle = angle_I - angle_U;
    if (*pAngle < -180.0f)
    {
      *pAngle += 360.0f;
    }
  }
  if (pRes)
  { 
    *pRes = amplitude_U / amplitude_I*1000 - 0.65;
  }
  
  /* apply calibration values */
  //printf("%1.5f %3.2f %1.5f %3.2f    ", amplitude_U, angle_U, amplitude_I, angle_I);
  //printf("%8.6f %3.2f\n", amplitude_U/amplitude_I*1000, angle_I - angle_U);
}

char *floatToMeterStr(double v, char unitchar, const char *prefix)
{
  if (v < 200.0e-12f)            /*    200p */
    sprintf(str2, "%3.1f p%c", v*1e12f, unitchar);
  else if (v < 2000.0e-12f)            /* 0-2000p */
    sprintf(str2, "%4.0f p%c", v*1e12f, unitchar);
  else if (v < 20.0e-9f)          /*     20n */
    sprintf(str2, "%2.2f n%c", v*1e9f, unitchar);
  else if (v < 200.0e-9f)         /*    200n */
    sprintf(str2, "%3.1f n%c", v*1e9f, unitchar);
  else if (v < 2.0e-6f)           /*      2u */
    sprintf(str2, "%1.3f u%c", v*1e6f, unitchar);
  else if (v < 20.0e-6f)          /*     20u */
    sprintf(str2, "%2.2f u%c", v*1e6f, unitchar);
  else if (v < 200.0e-6f)         /*    200u */
    sprintf(str2, "%3.1f u%c", v*1e6f, unitchar);
  else if (v < 2.0e-3f)           /*      2m */
    sprintf(str2, "%1.3f m%c", v*1e3f, unitchar);
  else if (v < 20.0e-3f)          /*     20m */
    sprintf(str2, "%2.2f m%c", v*1e3f, unitchar);
  else if (v < 200.0e-3f)         /*   200m */
    sprintf(str2, "%3.1f m%c", v*1e3f, unitchar);
  else if (v < 2.0f)              /*     2  */
    sprintf(str2, "%1.3f %c", v, unitchar);
  else if (v < 20.0f)             /*    20  */
    sprintf(str2, "%2.2f %c", v, unitchar);
  else if (v < 200.0f)            /*   200  */
    sprintf(str2, "%3.1f %c", v, unitchar);
  else if (v < 2.0e3f)            /*     2k */
    sprintf(str2, "%1.3f k%c", v*1e-3f, unitchar);
  else if (v < 20.0e3f)           /*    20k */
    sprintf(str2, "%2.2f k%c", v*1e-3f, unitchar);
  else if (v < 200.0e3f)          /*   200k */
    sprintf(str2, "%3.1f k%c", v*1e-3f, unitchar);
  else if (v < 2.0e6f)          /*       2M */
    sprintf(str2, "%1.3f M%c", v*1e-6f, unitchar);
  else
    sprintf(str2, "%2.2f M%c", v*1e-6f, unitchar);

  sprintf(str, "%s%s", prefix, str2);
  
  return str;
}




                   
int main( void )
{  
  uint32_t  f, freq;
  double Vp, Vq, Ip, Iq;
  double Rs, Xs, Cs, Ls, Zs, Qual;
  
  // Stop watchdog timer to prevent time out reset
  WDTCTL = WDTPW + WDTHOLD;
  
  sys_init();
  sys_extpower_switch(true);
  uart_init();
  pga_init();
  lcd_init();
  
  //lcd_print("Hello World!");  
  key_init();

  
  dac_init();
  dac_start_wave(FREQUENCY_100HZ);
  
//  printf("Hello World!\n");
  
  //lcd_printstring(0, 0, Impact_12, "1234567,890");

  calibrate_relative_gain();
  calibrate_cmrr();
  calibrate_gain();
  
   
  f = FREQUENCY_100HZ;
  while (1)
  {
    double res, angle;
    uint8_t key;
    
    
    key = get_key();
    if (key)
    {
      wait_key();
      f++;
      if (f > FREQUENCY_10000HZ)
      {
        f = FREQUENCY_100HZ;
      }
    }

    switch (f)
    {
    case FREQUENCY_100HZ:
      freq = 100;
      break;
    case FREQUENCY_1000HZ:
      freq = 1000;
      break;
    case FREQUENCY_10000HZ:
      freq = 10000;
      break;
    }
    breakoperation = false;
    measure_once(f, &res, &angle, &Vp, &Vq, &Ip, &Iq);
    if (!breakoperation)
    {
      //printf("%8.6f %3.2f    ", res, angle);    
      //printf("%8.6f %8.6f %8.6f %8.6f     ", Vp, Vq, Ip, Iq);    
      Vp = -Vp;
      Vq = -Vq;
      
      Rs = (Vp*Ip + Vq*Iq)/(Ip*Ip + Iq*Iq);
      Xs = (Vq*Ip - Vp*Iq)/(Ip*Ip + Iq*Iq);
      Cs = 1/(2*3.14159265f*freq*fabs(Xs));
      Ls = fabs(Xs)/(2*3.14159265f*freq);
      
      if (freq == 10000)
      { /* apply correction due to bad filtering of high frequencies */
        //Cs = Cs - 15e-12f;
        //if (Cs < 0)
        //  Cs = 0;
      }
  
      
      Zs = sqrt(Rs*Rs + Xs*Xs);
      Rs = Zs - fabs(Xs);
      if (Rs < 0)
        Rs = 0;
  
      Qual = fabs(Xs)/Rs;
      
      //printf("%8.6f %8.6f %8.6f %8.6f   ", Rs, Xs, Cs*1000000, Ls*1000000);    
      //printf("Q = %8.6f    %3.1f    ", Qual, angle);
      
      if (Zs > 5000000.0f)
      {
        sprintf(str, "---");    
        lcd_clear();
        lcd_printstring(0, 3, Impact_12, "---");
      }
      else
      {
        if (Qual < 0.15f)
        { /* pure resistive */
          lcd_clear();
          lcd_printstring(0, 3, Impact_12, floatToMeterStr(Rs, 'R', "R= "));
        }
        else if (Xs < 0)
        {
          lcd_clear();
          if (Qual > 1)
          {
            lcd_printstring(0, 2, Arial_7, floatToMeterStr(Rs, 'R', "R= "));
            lcd_printstring(0, 3, Impact_12, floatToMeterStr(Cs, 'F', "C= "));
          }
          else
          {
            lcd_printstring(0, 2, Impact_12, floatToMeterStr(Rs, 'R', "R= "));
            lcd_printstring(0, 4, Arial_7, floatToMeterStr(Cs, 'F', "C= "));
          }
        }
        else
        {
          lcd_clear();
          if (Qual > 1)
          {
            lcd_printstring(0, 2, Arial_7, floatToMeterStr(Rs, 'R', "R= "));
            lcd_printstring(0, 3, Impact_12, floatToMeterStr(Ls, 'H', "L= "));
          }
          else
          {
            lcd_printstring(0, 2, Impact_12, floatToMeterStr(Rs, 'R', "R= "));
            lcd_printstring(0, 4, Arial_7, floatToMeterStr(Ls, 'H', "L= "));
          }
        }
      }
      sprintf(str, "f = %ld Hz", freq);
      lcd_printstring(0, 0, Arial_7, str);
    }    
    //__delay_cycles(6000000ul);
  }

  return 0;
}
