/*
 * MSP430 LCR meter
 * Free for non-commercial use.
 * 
 * Contact: xyphro@gmail.com
 */
 
#include "measure.h"
#include "dac.h"
#include "sinetabs.h"
#include <math.h>
#include "io430.h"
#include <intrinsics.h>                                                         /* needed for __dleay_cycles macro */


#define ADC_MIN_VALUE 5
#define ADC_MAX_VALUE 4090


/* calculate mean, ADC min & max value*/
static inline void measure_step1(const int16_t *pinputsamples, 
                                 int16_t *pmean, int16_t *pmin, int16_t *pmax)
{
  int16_t value;
  int32_t mean, min, max;
  
  mean = 0;
  min = 4095;
  max = 0;
  for (int i=SAMPLE_COUNT-1; i > 0; i--)
  {
    value = *pinputsamples++;
    if (value > max)
      max = value;
    if (value < min)
      min = value;
    mean += value;
  }
  mean = (mean + (SAMPLE_COUNT>>1)) / 1200;
  
  *pmean = mean;
  *pmax = max;
  *pmin = min;
}

/* calculate a DFT of size 1200 for a single bin => faster as a FFT 
   as you see a discrete Fourier transform, so don't be scared by the name 
   basically it is just the mean value of the sample data which get shifted down 
   in frequency domain using a complex rotating vector :-) 
   The result is scaled by a factor of 0.95*1200 (0.95 due to the scaling in the 
   sine tables).
   fixed point data type for output = Q20
*/
static inline void measure_dft(const int16_t *pinputsamples, const int16_t *psinetab, uint16_t sin_tab_length, int16_t mean,
                 int32_t *presult_r, int32_t *presult_i)
{
  const int16_t *pcur_twid_i;                                                   /* pointer to current IMAGINARY twiddle factor */
  const int16_t *pcur_twid_r;                                                   /* pointer to current REAL twiddle factor */
  const int16_t *pwrap_ptr_position;                                            /* pointer used for wrap over checks of twiddle factors */
  int16_t value;
  int32_t res_r, res_i;
  int16_t twid_r, twid_i;                                                       /* complex twiddle factors */
  //int32_t res_tmp;
  
  pcur_twid_i = psinetab;
  pcur_twid_r = pcur_twid_i + (sin_tab_length>>2);                              /* real twiddles are just shifted by 90 */
  pwrap_ptr_position = psinetab + sin_tab_length;                               /* this points to undefined memory, but it is only used for comparision */
  res_r = 0;
  res_i = 0;
  //MPY32CTL0 = MPYM_1;   /* signed multiplication, 16 bit operands, no sat, no frac */
  for (int i=0; i < SAMPLE_COUNT-1; i++)
  {
    value = *pinputsamples++;                                                   /* get current sample from ADC sample table */
    value -= mean;                                                              /* not really needed... */
    twid_r = *pcur_twid_r++;
    twid_i = *pcur_twid_i++;                                                    /* get twiddle factors */
    
    /* TODO: This can be speed up by using fractional modes of the hardware multiplier */
    
    /*MPYS = value;
    OP2 = twid_r;
    __asm("nop");__asm("nop");__asm("nop");
    res_tmp = RESHI;                           // Load RESHI word result
    res_r += (res_tmp<<16)|RESLO;              // Shift RESHI left and concat with
    */
    res_r += ((int32_t)value * (int32_t)twid_r) >> 2; /* (Q11 * Q11)>>2 = Q22>>2 = Q20 */

    //MPYS = value;
    /*OP2 = twid_i;
    __asm("nop");__asm("nop");__asm("nop");
    res_tmp = RESHI;                           // Load RESHI word result
    res_i += (res_tmp<<16)|RESLO;              // Shift RESHI left and concat with
    */
    res_i += ((int32_t)value * (int32_t)twid_i) >> 2; /* (Q11 * Q11)>>2 = Q22>>2 = Q20 */
    
    /* check if twiddle ptrs need to be wraped over to start of table */
    if (pcur_twid_i == pwrap_ptr_position)
      pcur_twid_i = psinetab;
    if (pcur_twid_r == pwrap_ptr_position)
      pcur_twid_r = psinetab;
  }
  
  *presult_r = res_r;
  *presult_i = res_i;
}




/**
 * measure
 * \param [in]          pinputsamples
 * \param [out]         over_underflow          0               no adc over/underflow
 *                                              1               adc overflow
 *                                              -1              adc underflow
 */
void measure(const int16_t *pinputsamples, uint8_t freq,                        /* input parameters */
             int16_t *pmean, int16_t *prough_amplitude, double *pangle, double *pamplitude, 
             uint8_t *pover_underflow, double *pamplitude_r, double *pamplitude_i)   /* output parameters */
{
  int16_t  mean, min_v, max_v;
  int8_t   over_underflow;
  int16_t  tmp1, tmp2;
  
  int16_t const *sin_tab;
  uint16_t sin_tab_length;  
  int32_t result_r, result_i;
  double amplitude;
  
  /* 1.) measure mean value and detect over/underflow */
  measure_step1(pinputsamples, &mean, &min_v, &max_v);
  /* check if over or underflow occured */
  over_underflow = MEASURE_SCALINGOK;
  if (min_v < ADC_MIN_VALUE)
    over_underflow |= MEASURE_UNDERFLOW;
  if (max_v > ADC_MAX_VALUE)
    over_underflow |= MEASURE_OVERFLOW;
  
  if (pmean)
    *pmean = mean;
  if (pover_underflow)
    *pover_underflow = over_underflow;  
  
  if (prough_amplitude)
  {
    tmp1 = min_v-mean;
    if (tmp1 < 0)
      tmp1 = -tmp1;
    tmp2 = max_v-mean;
    if (tmp2 < 0)
      tmp2 = -tmp2;
    if (tmp1 > tmp2)
      *prough_amplitude = tmp1;
    else
      *prough_amplitude = tmp2;
  }
  
  if (over_underflow)                                                           /* stop processing herein case of over or underflow */
    return;
  if (!pamplitude && !pangle)                                                   /* stop if dft calculation is not really needed */
    return;
  
  /* 2.) select twiddle table and bin for the following DFT calculation */
  switch (freq)
  {
  case FREQUENCY_100HZ:
    sin_tab = Sin_tab_1kHz;
    sin_tab_length = sizeof(Sin_tab_1kHz) >> 1;
    break;
  case FREQUENCY_1000HZ:
  case FREQUENCY_1000HZ_CAL:    
    sin_tab = Sin_tab_10kHz;    
    sin_tab_length = sizeof(Sin_tab_10kHz) >> 1;
    break;
  case FREQUENCY_10000HZ:
  default:
    sin_tab = Sin_tab_100kHz;
    sin_tab_length = sizeof(Sin_tab_100kHz) >> 1;
    break;
  }
  
  /* 3.) calculate DFT to get phase and amplitude */
  measure_dft(pinputsamples, sin_tab, sin_tab_length, mean, &result_r, &result_i);
  if (pamplitude_r)
  {
    *pamplitude_r = result_r;
  }
  if (pamplitude_i)
  {
    *pamplitude_i = result_i;
  }
  if (pamplitude)
  {
    amplitude = sqrt( (double)result_r*(double)result_r + (double)result_i*(double)result_i );
    amplitude = amplitude / (0.95f*1200.0f*524288.0f); /* normalize amplitude */
    *pamplitude = amplitude;
  }
  if (pangle)
    *pangle = atan2((double)result_r, (double)result_i)/3.14159265f * 180.0f; 
 
}






/* TRUNC - functions before core specific optimizations */
#if 0
/* NON-OPZIMIZED, because it does not make efficient use of hardware multipliers, etc... */
static inline void measure_dft(const int16_t *pinputsamples, const int16_t *psinetab, uint16_t sin_tab_length, int16_t mean,
                 int32_t *presult_r, int32_t *presult_i)
{
  const int16_t *pcur_twid_i;                                                   /* pointer to current IMAGINARY twiddle factor */
  const int16_t *pcur_twid_r;                                                   /* pointer to current REAL twiddle factor */
  const int16_t *pwrap_ptr_position;                                            /* pointer used for wrap over checks of twiddle factors */
  int16_t value, tmp;
  int32_t res_r, res_i;
  int16_t twid_r, twid_i;                                                       /* complex twiddle factors */
  
  pcur_twid_i = psinetab;
  pcur_twid_r = pcur_twid_i + (sin_tab_length>>2);                              /* real twiddles are just shifted by 90 */
  pwrap_ptr_position = psinetab + sin_tab_length;                               /* this points to undefined memory, but it is only used for comparision */
  res_r = 0;
  res_i = 0;
  for (int i=0; i < SAMPLE_COUNT-1; i++)
  {
    value = *pinputsamples++;                                                   /* get current sample from ADC sample table */
    value -= mean;                                                              /* not really needed... */
    twid_r = *pcur_twid_r++;
    twid_i = *pcur_twid_i++;                                                    /* get twiddle factors */
    
    /* TODO: This can be speed up by using fractional modes of the hardware multiplier */
    res_r += ((int32_t)value * (int32_t)twid_r) >> 2; /* (Q11 * Q11)>>2 = Q22>>2 = Q20 */
    res_i += ((int32_t)value * (int32_t)twid_i) >> 2; /* (Q11 * Q11)>>2 = Q22>>2 = Q20 */
    
    /* check if twiddle ptrs need to be wraped over to start of table */
    if (pcur_twid_i == pwrap_ptr_position)
      pcur_twid_i = psinetab;
    if (pcur_twid_r == pwrap_ptr_position)
      pcur_twid_r = psinetab;
  }
  
  *presult_r = res_r;
  *presult_i = res_i;
}
#endif